// Copyright (c) 2016, Tobias Mueller tm(at)tm3d.de
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the
//    distribution.
//  * All advertising materials mentioning features or use of this
//    software must display the following acknowledgement: This product
//    includes software developed by tm3d.de and its contributors.
//  * Neither the name of tm3d.de nor the names of its contributors may
//    be used to endorse or promote products derived from this software
//    without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "owPiGPioInterface.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/mman.h>
#include <fcntl.h>    /* For O_RDWR */
#include <unistd.h>
#include <pthread.h>
#include <sched.h>
#include <memory.h>
#include <strings.h>

namespace gpio_1w {


#ifndef MAP_LOCKED
#define MAP_LOCKED 0
#endif
static volatile uint32_t pi_peri_phys =0;


#define GPIO_BASE  (pi_peri_phys + 0x00200000)
#define SYST_BASE  (pi_peri_phys + 0x00003000)

#define GPIO_LEN  0xB4
#define SYST_LEN  0x1C

#define USD(ms) {unsigned int t=TICK+ms;while (TICK<t);}


static int fdMem        = -1;

static volatile uint32_t * gpioReg = 0;
static volatile uint32_t * systReg = 0;

static volatile uint32_t * gpioReg_read = 0;




int gpio_reg;
int gpio_shift;
int gpio_shift_;
int gpio_mode_shift;

#define OW_SET_LOW gpioReg[gpio_reg] = (gpioReg[gpio_reg] & gpio_shift_) | (gpio_mode_shift);
#define OW_RELEASE_LOW gpioReg[gpio_reg] = (gpioReg[gpio_reg] & gpio_shift_) ;

#define SYST_CLO     1
#define TICK systReg[SYST_CLO]
 

#define GPLEV0    13
#define GPCLR0    10
#define GPSET0     7

int BANK=0;//=(owpin>>5);
int BIT=0;//=(1<<(owpin&0x1F));

#define OW_READ ((*(gpioReg_read) & BIT) != 0) 

int piCores;

unsigned PIHardwareRevision(void)
{
   static unsigned rev = 0;

   FILE * filp;
   char buf[512];
   char term;

   if (rev) return rev;

   piCores = 0;

   filp = fopen ("/proc/cpuinfo", "r");

   if (filp != NULL)
   {
      while (fgets(buf, sizeof(buf), filp) != NULL)
      {
         if (piCores == 0)
         {
            if (!strncasecmp("model name", buf, 10))
            {
               if (strstr (buf, "ARMv6") != NULL)
               {
                  piCores = 1;
                  pi_peri_phys = 0x20000000;
            //      pi_dram_bus  = 0x40000000;
             //     pi_mem_flag  = 0x0C;
               }
               else if (strstr (buf, "ARMv7") != NULL)
               {
                  piCores = 4;
                  pi_peri_phys = 0x3F000000;
                  //pi_dram_bus  = 0xC0000000;
                  //pi_mem_flag  = 0x04;
               }
               else if (strstr (buf, "ARMv8") != NULL)
               {
                  piCores = 4;
                  pi_peri_phys = 0x3F000000;
               //   pi_dram_bus  = 0xC0000000;
                //  pi_mem_flag  = 0x04;
               }
            }
         }

         if (!strncasecmp("revision\t:", buf, 10))
         {
            if (sscanf(buf+10, "%x%c", &rev, &term) == 2)
            {
               if (term != '\n') rev = 0;
               else rev &= 0xFFFFFF; /* mask out warranty bit */
            }
         }
      }

      fclose(filp);
   }
   return rev;
}






}
using namespace gpio_1w;


int owPiGPioInterface::InitAdapter(uint8_t pin){
    owpin=pin;
    PIHardwareRevision();
    if (fdMem==-1) {
		if ((fdMem = open("/dev/mem", O_RDWR | O_SYNC) ) < 0) {
			log->set(OWLOG_ERROR,"No permission to change Port States, try running as root (sudo)");
			exit(-1);
		}
	
		systReg  = (uint32_t *) mmap(0, SYST_LEN,  PROT_READ|PROT_WRITE|PROT_EXEC,
										MAP_SHARED|MAP_LOCKED,	fdMem, SYST_BASE);
		gpioReg  = (uint32_t *) mmap(0, GPIO_LEN,  PROT_READ|PROT_WRITE|PROT_EXEC,
										MAP_SHARED|MAP_LOCKED,	fdMem, GPIO_BASE);
		//if (level == PI_OFF) *(gpioReg + GPCLR0 + BANK) = BIT;
		//else                 *(gpioReg + GPSET0 + BANK) = BIT;
		*(gpioReg + GPCLR0 + BANK) = BIT;
	
		gpioReg_read=gpioReg + GPLEV0 + BANK;
	
		gpio_reg=owpin/10;
		gpio_shift= (owpin%10) * 3;
		gpio_shift_=~(7<<gpio_shift);
		gpio_mode_shift=1<<gpio_shift;
	
		BANK=(owpin>>5);
		BIT=(1<<(owpin&0x1F));
	}
	return 0;	
	
}
int owPiGPioInterface::Reset() {
	int read;
	OW_SET_LOW;
	USD(480);
	OW_RELEASE_LOW;
	USD(65);
	read=OW_READ;
	USD(410);
	if (read) {
		OW_SET_LOW;
		USD(480);
                OW_RELEASE_LOW;
                USD(65);
		read=OW_READ;
		USD(410);

	}
	return read==0;
}


uint8_t owPiGPioInterface::sendrecivBit(uint8_t bit) {
	unsigned int lmis;
	uint8_t read=0;
	//pinMode(owpin,OUTPUT);
  	lmis=TICK;
  	
  		//cli_();
	OW_SET_LOW;
	lmis+=6;
	while (TICK<lmis) ;
	if (bit) {
		OW_RELEASE_LOW;
		lmis+=11;
		while (TICK<lmis) ;
		read=OW_READ;
		while ((TICK-lmis)<49);
		
	} else {
		while ((TICK-lmis)<60);
		OW_RELEASE_LOW;
	}
	//sti();
	USD(10);
	return read;
}



